#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <sys/mman.h>

/****************
 ** This command will reset the screen mode after testing the mode.
 ** You should change this!
 */
#define FBRESETCOMMAND "fbset -g 1280 1024 1504 1024 8 -t 9292 232 64 38 1 112 3 -accel true"


/* This module does the mode previewing in framebuffer console. */

#ifdef DJGPP
#include <crt0.h>
#include <conio.h>
#else
#include <termio.h>
#include <termios.h>
#include <sys/ioctl.h>
#endif

#include "preview.h"

static unsigned LINES, COLS;

static void GetScreenGeometry(void)
{
#if defined(DJGPP) || defined(__BORLANDC__)
    struct text_info w;
    memset(&w, 0, sizeof(w));
    gettextinfo(&w);
    LINES=w.screenheight;
    COLS =w.screenwidth;
#else
#ifdef TIOCGWINSZ
    struct winsize w;
    memset(&w, 0, sizeof(w));
    if(ioctl(1, TIOCGWINSZ, &w) >= 0)
    {
        LINES=w.ws_row;
        COLS =w.ws_col;
    }
    else
#else
#ifdef WIOCGETD
    struct uwdata w;
    memset(&w, 0, sizeof(w));
    LINES = COLS = 0;
    if(ioctl(1, WIOCGETD, &w) >= 0)
    {
        LINES = w.uw_height / w.uw_vs;
        COLS  = w.uw_width / w.uw_hs;
    }
    else
#endif
#endif
    LINES = COLS = 0;
#endif
}

static void PutC(unsigned char *map, unsigned x, unsigned y, unsigned colour, unsigned mrk)
{
    map[(y*COLS+x)*2]   = mrk;
    map[(y*COLS+x)*2+1] = colour;
}

void FillScreen(void)
{
    FILE *fp = fopen("/dev/vcsa0", "rb+");
    unsigned char *map;
    unsigned x, y, xma, yma;
    int hasmmap = 0;
    
    if(!fp)return;
    
    GetScreenGeometry();
    
    if(!LINES || !COLS)
    {
        fclose(fp);
        return;
    }

    map = mmap(NULL, LINES*COLS*2, PROT_WRITE, MAP_PRIVATE, fileno(fp), 4);
    
    if(map == MAP_FAILED)
    {
        /* Actually, the mmap *always* fails for vcs-devices :( */
        map = malloc(LINES*COLS*2);
        if(!map)
        {
            fclose(fp);
            return;
        }
    }
    else
    {
        hasmmap = 1;
        fclose(fp);
    }
    
    xma = COLS/6;  if(xma < 4)xma = 4;
    yma = LINES/6; if(yma < 4)yma = 4;
    
    for(y=0; y<LINES; ++y)
        for(x=0; x<COLS; ++x)
            PutC(map, x,y, 0x07, (unsigned char) ("PRESS A KEY  "[x%(6+2+3+2)]));

    for(y=0; y<yma; ++y)
        for(x=0; x<xma; ++x)
        {
            PutC(map, x,       y,         0x1F, '0'+(y>x?y:x)%10);
            PutC(map, COLS-x-1,y,         0x4F, '0'+(y>x?y:x)%10);
            PutC(map, COLS-x-1,LINES-y-1, 0x2E, '0'+(y>x?y:x)%10);
            PutC(map, x,       LINES-y-1, 0xD0, '0'+(y>x?y:x)%10);
        }
    
    if(hasmmap)
    {
        munmap(map, LINES*COLS*2);
    }
    else
    {
        fseek(fp, 4, SEEK_SET);
        fwrite(map, 2, LINES*COLS, fp);
        fclose(fp);
    }

    printf("\33[%d;%dH", LINES, 1);
}

void SetMode(void)
{
    char Buf[512];
    
    sprintf(Buf,
        "fbset -g %u %u %u %u 8 -t %u %u %u %u %u %u %u -vsync high -hsync high",
        xres, yres,
        xres, yres,
        (unsigned) (1e12 / dispclock),
        left, right,
        top, bottom,
        hslen, vslen
    );
    
    fprintf(stderr, "Executing: '%s'\n", Buf);
    system(Buf);
    GetScreenGeometry();
    printf("\33[%d;%dH", LINES-10, 1);
}

void RestoreMode(void)
{
    /* Restore the videomode to normal */
    system(FBRESETCOMMAND);
}
